
__version__ = "10.4.0"
__author__ = "Alex Al-Saffar"
__date__ = "2023-02-24"
docs = r"https://crocodile.readthedocs.io/en/latest/"
git = r"https://github.com/thisismygitrepo/crocodile"
pypi = r"https://pypi.org/project/crocodile"
release_notes = """
added bzip2 compression
"""

